//Copyright (c) 2000-2003  San Diego Supercomputer Center (SDSC),
//a facility operated by the University of California, San Diego (UCSD)
//
//Users and possessors of this source code are hereby granted a
//nonexclusive, royalty-free copyright and design patent license to
//use this code in individual software.  License is not granted for
//commercial resale, in whole or in part, without prior written
//permission from SDSC.  This source is provided "AS IS" without express
//or implied warranty of any kind.
//
//For further information, please see:  http://mbt.sdsc.edu
//
package edu.sdsc.sirius.io;


import edu.sdsc.mbt.*;
import edu.sdsc.mbt.util.*;

import java.util.*;
import java.io.*;
import java.net.*;

public class GamessEntryLoader implements FileEntryLoader, java.io.Serializable {

	protected String urlString;
	private long expectedInputBytes = 1;

	private final int DATA_TYPE = Entry.DATA_STRUCTURE;

	// A hashtable of vectors where
	// each hash KEY is the StructureComponent type String.
	// each hash VALUE is a Vector of StructureComponent objects.
	// NOTE: This protected variable is used only to enable each new
	// Structure object to access the newly loaded data. The Structure
	// constructor stores a copy of the object so this variable can
	// be re-used for the next load request. This is not thread-safe
	// but there doesn't seem to be any other way to enable an
	// anonymous class to be handed state (aside from parameters).
	protected Hashtable passComponents;


	//
	// StructureLoader interface methods
	//
	/**
	 * Returns name of the loader.
	 */
	public String getLoaderName(){
		return new String("GAMESS Cartesian Input Format");
	}

	/**
	 * Loads the Entry by file name.
	 */
	public Entry load( String name, int number ){
	
		StructureEntry entry = new StructureEntry();
		
		Structure structure = null;//future structureObject derived from the PDB file
		try
		{	
			File file = new File( name );
			structure = loadStructure( file );
		}
		catch( NullPointerException e )
		{
			try
			{	
				URL url = new URL( name );
				structure = loadStructure( url, null );
				entry.setSourceUrl(url);
			}
			catch( MalformedURLException e2 )
			{
			}
		}
		if ( structure == null ){
			return null;
		}
		
		urlString = name;
		
		//modify the file name to get rid of the extension and use this string as the entry name
		String shortName = name;
		if (name.endsWith(".inp") || name.endsWith(".INP")){
			shortName = name.substring(0, name.length()-4);
		}
		
		if(shortName.indexOf(File.separator) > -1){
			int index = shortName.lastIndexOf(File.separator);
			String temp = shortName.substring(index+1);
			shortName = temp;
		}
		
		entry.setStructure(structure);
		
		entry.setName(shortName);
		
		return entry;
	}
	
	/**
	 * Loads the Entry by file name and sets all identificators to the specified String.
	 */
	public Entry load( String name, String identificator){
	
		StructureEntry entry = new StructureEntry();
		
		Structure structure = null;//future structureObject derived from the PDB file
		try
		{	
			File file = new File( name );
			structure = loadStructure( file );
		}
		catch( NullPointerException e )
		{
			try
			{	
				URL url = new URL( name );
				structure = loadStructure( url, identificator );
				entry.setSourceUrl(url);
			}
			catch( MalformedURLException e2 )
			{
			}
		}
		if ( structure == null ){
			return null;
		}
		
		urlString = identificator;
		
		//modify the file name to get rid of the extension and use this string as the entry name
		String shortName = identificator;
		if (identificator.endsWith(".inp") || identificator.endsWith(".INP")){
			shortName = identificator.substring(0, identificator.length()-4);
		}
		
		if(shortName.indexOf(File.separator) > -1){
			int index = shortName.lastIndexOf(File.separator);
			String temp = shortName.substring(index+1);
			shortName = temp;
		}

		entry.setStructure(structure);		
		entry.setName(shortName);
		
		return entry;
	}

	
	
	/**
	 * Loads the Entry from a File object.
	 */
	public Entry load(File file, int number){
	
		StructureEntry entry = new StructureEntry();
		
		String name = file.toString();
//		urlString = name;
		
		Structure structure = null;
		try {	
			structure = loadStructure( file );
			entry.setSourceUrl(file.toURL());
		}
		catch (NumberFormatException ex){
			throw ex;
		}
		catch( Exception e ){
			e.printStackTrace();
		}
		
		if ( structure == null ){
			return null;
		}
		
		//modify the file name to get rid of the extension and use this string as the entry name
		String shortName = name;
		if (name.endsWith(".inp") || name.endsWith(".INP")){
			shortName = name.substring(0, name.length()-4);
		}

		if(shortName.indexOf(File.separator) > -1){
			int index = shortName.lastIndexOf(File.separator);
			String temp = shortName.substring(index+1);
			shortName = temp;
		}

		entry.setStructure(structure);
		entry.setName(shortName);
		
		return entry;
	
	
	}
	

	
	/**
	 * Loads the Entry from a URL object.
	 */
	public Entry load(URL url, int number){
	
		StructureEntry entry = new StructureEntry();
		
		String name = url.toString();
		
		Structure structure = loadStructure( url, null );

		if ( structure == null ){
			return null;
		}
		
		entry.setSourceUrl(url);
		
		//modify the file name to get rid of the extension and use this string as the entry name
		String shortName = name;
		if (name.endsWith(".inp") || name.endsWith(".INP")){
			shortName = name.substring(0, name.length()-4);
		}

		if(shortName.indexOf(File.separator) > -1){
			int index = shortName.lastIndexOf(File.separator);
			String temp = shortName.substring(index+1);
			shortName = temp;
		}

		entry.setStructure(structure);
		entry.setName(shortName);
		
		return entry;
	}

	
	/**
	 * Returns true if the loader is capable of loading the structure,
	 * or false otherwise. This enables higher-level code to be able
	 * to build a context sensative menu of only the loaders that can
	 * load a given structure name.
	 */
	public boolean canLoad( String name )
	{
		// System.err.println( "PdbStructureLoader.canLoad(String)" );
		if ( ! name.endsWith( ".inp" ) && ! name.endsWith(".INP"))
			return false;

		return true;
	}

	//
	// FileStructureLoader interface methods
	//
	
	/**
	 * Returns a reference to a Structure read from the given File object.
	 */
	public Structure loadStructure( File file )
	{
		// System.err.println( "PdbStructureLoader.load(File)" );
		
//		System.out.println("loadStructure with id = " + identificator);

		if ( ! canLoad( file ) ) return null;
		if ( ! file.exists() ) return null;
		if ( ! file.canRead() ) return null;

		expectedInputBytes = file.length( );
		if ( expectedInputBytes <= 0 ) return null;

		try
		{
			FileInputStream fileInputStream = new FileInputStream( file );
			if ( fileInputStream == null ) return null;
			
			//filename was being reduced to just the name itself here and in SequenceViewerPanel
//			urlString = file.toURL().toExternalForm();
			
			String tempUrlString = file.toURL().toExternalForm();
				
			//process the name to get rid of the full path and keep
			//only the file name without the extension for internal reference purposes
			int index = tempUrlString.lastIndexOf("/") + 1;
			String tUrlString = tempUrlString.substring(index);
		
			index = tUrlString.lastIndexOf(".");
			urlString = tUrlString.substring(0,index);
			
						
			return loadStructure( fileInputStream );
		}
		catch (NumberFormatException e){
			throw e;
		}
		catch ( FileNotFoundException e )
		{
			return null;
		}
		catch ( MalformedURLException e )
		{
			return null;
		}
		catch ( SecurityException e )
		{
			return null;
		}
	}


	/**
	 * Returns true if the Entry can be read from the given File object.
	 */
	public boolean canLoad( File file )
	{
		// System.err.println( "PdbStructureLoader.canLoad(File)" );
		if ( ! file.exists() ) return false;

		try
		{
			// String path = file.getCanonicalPath( );
			String path = file.toURL().toExternalForm();
			return canLoad( path );
		}
		catch( IOException e )
		{
			return false;
		}
	}


	/**
	 * Returns a reference to a Structure read from the given URL object.
	 */
	public Structure loadStructure( URL url, String identificator )
	{
		// System.err.println( "PdbStructureLoader.load(URL)" );

		if ( ! canLoad( url ) ) return null;

		try
		{
			URLConnection urlConnection = url.openConnection( );
			expectedInputBytes = urlConnection.getContentLength( );
			if ( expectedInputBytes <= 0 ) return null;
			InputStream inputStream = urlConnection.getInputStream( );
			
			if (identificator != null){
				urlString = identificator;
			}

			if ( inputStream != null ) urlString = url.toExternalForm( );
			
			return loadStructure( inputStream );
		}
		catch( IOException e )
		{
			return null;
		}
	}


	/**
	 * Returns true if the Entry can be read from the given URL object.
	 */
	public boolean canLoad( URL url )
	{
		// System.err.println( "PdbStructureLoader.canLoad(URL)" );
		return canLoad( url.toExternalForm( ) );
	}

	//
	// PdbStructureLoader methods
	//

	/**
	 * Returns a reference to a Structure read from the given (uncompressed)
	 * InputStream.
	 */
	public Structure loadStructure( InputStream bufferedInputStream ){
		
		if ( bufferedInputStream == null ) return null;

		passComponents = new Hashtable( );
		long expectedBytes = expectedInputBytes;
		
		byte buf[] = new byte[ 2048 ]; // raw read data buffer
		byte line[] = new byte[ 100 ]; // accumulated line buffer

		int expectedReads = (int) ((float) expectedBytes / (float) buf.length);
		int onePercent = (int) (((float) expectedBytes / (float) buf.length) / 100.f);
		if ( onePercent <= 0 ) onePercent = 1;
		float percentDone = 0.0f;

		Status.progress( percentDone, "Chargement " + urlString );
		
		int linePos = 0;
		int bytesRead = 0;
		int lastRead = 0;
		int lines = 0;
		int readCount = 0;
		
		String name = null;
		
		Atom atom = null;
		int atomCounter = 1;
		
		int startLine = -1;
		boolean readData = false;
		
		int atomNumber = 1;
		
		try
		{
			while ( bufferedInputStream.available() > 0 )
			{
				// Re-fill the raw data buffer.
				try
				{
					lastRead = bufferedInputStream.read( buf, 0, buf.length );
				}
				catch( IOException e )
				{
//					System.err.println( "XMLLoader.load: " + e );
					return null;
				}
				bytesRead += lastRead;

				readCount++;
				if ( readCount % onePercent == 0 )
				{
					percentDone = (float) bytesRead / (float) expectedBytes;
					Status.progress( percentDone, "Chargement " + urlString );
				}
	
				// Process the buffer
				for ( int bufPos=0; bufPos<lastRead; bufPos++ )
				{
					// Copy a byte from the input buffer into the line buffer
					line[linePos] = buf[bufPos];
					if ( buf[bufPos] == '\n' )
					{
						// line[0-linePos] now has a complete line.
						lines++;
						
						String data = new String(line, 0, linePos);
//						System.out.println("atom = " + atom + ", data = #" + data + "#");
						
						if (data.indexOf("$DATA") > -1){
							startLine = lines;
							readData = true;
							linePos = 0;
							continue;
						}
						
						if (readData && lines == (startLine+1)){
							//the content is name
							urlString = data.substring(0, data.length()-1);
//							System.out.println("lines = " + lines + ", name = " + urlString);
							linePos = 0;
							continue;
						}
						
						if (readData && lines == (startLine+2)){
							//this is a symmetry line
							linePos = 0;
							readData = true;
							continue;
						}
						
						if (readData && data.indexOf("$END") > -1){
							readData = false;
							break;
						}
						
						if (readData){
//							System.out.println("Creating new atom");
							atom = new Atom();
							atom.coordinate = new double[3];
							atom.chain_id = "_";

							atom.number = atomNumber++;
							atom.compound = "RES";
							
							
							
							int counter = 0;
							StringTokenizer tok = new StringTokenizer(data, " ", false);
							while (tok.hasMoreTokens()){
								String token = tok.nextToken();
								
//								System.out.println("token = " + token);
								
								if (counter == 0){
									
									if (token.length() > 1){
										atom.setElement(token.substring(0, 1));
									}
									else{
										atom.setElement(token);
									}
								}
								
								if (counter == 1){
									//get the atomic number
									int n = -1;
									try{
										n = Integer.parseInt(token);
									}
									catch (NumberFormatException e){
										try{
											n = (int)Float.parseFloat(token);
										}
										catch (NumberFormatException x){
											x.printStackTrace();
										}
									}
									atom.setElement(PeriodicTable.getElementSymbol(n));
									atom.name = atom.getElement() + atomCounter;
									atomCounter++;
									
								}
								
								if (counter == 2){
									//x
									try{
										atom.coordinate[0] = Double.parseDouble(token);
									}
									catch (NumberFormatException ee){
										Status.progress(1.0f, null);
										return null;
//										System.out.println("failure for " + token);
//										ee.printStackTrace();
									}
								}
								
								if (counter == 3){
									//x
									try{
										atom.coordinate[1] = Double.parseDouble(token);
									}
									catch (NumberFormatException ee){
										Status.progress(1.0f, null);
										return null;
//										System.out.println("failure for " + token);
//										ee.printStackTrace();
									}
								}
							
								if (counter == 4){
									//x
									try{
										atom.coordinate[2] = Double.parseDouble(token);
									}
									catch (NumberFormatException ee){
										Status.progress(1.0f, null);
										return null;
//										System.out.println("failure for " + token);
//										ee.printStackTrace();
									}
								}
								

								counter++;
							}
							
//							System.out.println("atom = " + atom.name + ", coordinates = " + atom.coordinate[0] + "," + atom.coordinate[1] + "," + atom.coordinate[2]);
							
							
							Vector records = (Vector) passComponents.get( StructureComponentRegistry.TYPE_ATOM );
							if ( records == null )	{
								records = new Vector( );
								passComponents.put( StructureComponentRegistry.TYPE_ATOM, records );
							}
							records.add( atom );

						
						}
						
						linePos = 0;
					}
					else
					{
						// Continue building the line.
						linePos++;
						if ( linePos >= line.length )
						{
//							System.err.println( "line buffer overflow!" );
							Status.progress( 1.0f, null );
							return null;
						}
					}
				}
			}
		}
		catch (Exception ex){
			ex.printStackTrace();
			Status.progress( 1.0f, null );
			return null;
		}
						
						
		Status.progress( 1.0f, null );
				
						
						
						
		
		//
		// Create the Structure object
		//
		Structure structure = new Structure()
		{
			// A hashtable of vectors where
			// each hash KEY is the StructureComponent type String.
			// each hash VALUE is a Vector of StructureComponent objects.
			protected Hashtable structureComponents = null;

			// To free up the global state for another load call.
			private String localUrlString;
			

			// public Structure()  Anonymous inner class constructor.
			{
				structureComponents = passComponents;
				passComponents = null;

				localUrlString = urlString;
				urlString = null;
				
			}

			public String getUrlString( ){
				return localUrlString;
			}

			
			public int getStructureComponentCount( String scType ){
				Vector records = (Vector) structureComponents.get( scType );
				if ( records == null )
					return 0;
				else
					return records.size( );
			}

			private void getStructureComponentByIndex( int index,
				StructureComponent structureComponent )
				throws IndexOutOfBoundsException, IllegalArgumentException
			{
				String type = structureComponent.getStructureComponentType();
				if ( type == null )
					throw new IllegalArgumentException( "type inconnu" + type );

				Vector records = (Vector) structureComponents.get( type );
				if ( records == null )
					throw new IllegalArgumentException( "pas d\'enregistrement du type " + type );

				StructureComponent cachedComponent =
					(StructureComponent) records.elementAt( index );

				cachedComponent.structure = this;
				structureComponent.structure = this;
				structureComponent.copy( cachedComponent );
            }

			public StructureComponent getStructureComponentByIndex( String type,
				int index )
				throws IndexOutOfBoundsException, IllegalArgumentException
			{
				Vector records = (Vector) structureComponents.get( type );
				if ( records == null )
					throw new IllegalArgumentException( "pas d\'enregistrement du type " + type );

				StructureComponent structureComponent =
					(StructureComponent) records.elementAt( index );
				structureComponent.structure = this;

				return structureComponent;
            }
		};

		// Progress is done.
		Status.progress( 1.0f, null );

		return structure;
	}
	
	public int getDataType(){
		return DATA_TYPE;
	}

		
}

