//Copyright (c) 2000-2003  San Diego Supercomputer Center (SDSC),
//a facility operated by the University of California, San Diego (UCSD)
//
//Users and possessors of this source code are hereby granted a
//nonexclusive, royalty-free copyright and design patent license to
//use this code in individual software.  License is not granted for
//commercial resale, in whole or in part, without prior written
//permission from SDSC.  This source is provided "AS IS" without express
//or implied warranty of any kind.
//
//For further information, please see:  http://mbt.sdsc.edu
//
package edu.sdsc.sirius.io;


import edu.sdsc.mbt.*;
import edu.sdsc.mbt.util.*;

import java.util.*;
import java.io.*;
import java.net.*;

public class Mol2EntryLoader implements FileEntryLoader, java.io.Serializable {

	protected String urlString = null;
	private long expectedInputBytes = 1;
	
	private final int DATA_TYPE = Entry.DATA_STRUCTURE;

	// A hashtable of vectors where
	// each hash KEY is the StructureComponent type String.
	// each hash VALUE is a Vector of StructureComponent objects.
	// NOTE: This protected variable is used only to enable each new
	// Structure object to access the newly loaded data. The Structure
	// constructor stores a copy of the object so this variable can
	// be re-used for the next load request. This is not thread-safe
	// but there doesn't seem to be any other way to enable an
	// anonymous class to be handed state (aside from parameters).
	protected Hashtable passComponents = null;


	//
	// StructureLoader interface methods
	//
	/**
	 * Returns name of the loader.
	 */
	public String getLoaderName(){
		return new String("format MOL2");
	}

	/**
	 * Loads the Entry by file name.
	 */
	public Entry load( String name, int number ){
	
		StructureEntry entry = new StructureEntry();
		
		Structure structure = null;//future structureObject derived from the PDB file
		try
		{	
			File file = new File( name );
			structure = loadStructure( file );
		}
		catch( NullPointerException e )
		{
			try
			{	
				URL url = new URL( name );
				structure = loadStructure( url );
				entry.setSourceUrl(url);
			}
			catch( MalformedURLException e2 )
			{
			}
		}
		if ( structure == null ){
			return null;
		}
		
		urlString = name;
		
		//modify the file name to get rid of the extension and use this string as the entry name
		String shortName = name;
		if (name.endsWith(".mol2")){
			shortName = name.substring(0, name.length()-4);
		}
		
		if(shortName.indexOf(File.separator) > -1){
			int index = shortName.lastIndexOf(File.separator);
			String temp = shortName.substring(index+1);
			shortName = temp;
		}
		
		entry.setStructure(structure);
		
		entry.setName(shortName);
		
		return entry;
	}
	
	
	/**
	 * Loads the Entry from a File object.
	 */
	public Entry load(File file, int number){
	
		StructureEntry entry = new StructureEntry();
		
		String name = file.toString();
//		urlString = name;
		
		Structure structure = null;
		try {	
			structure = loadStructure( file );
			entry.setSourceUrl(file.toURL());
		}
		catch( Exception e ){}
		
		if ( structure == null ){
			return null;
		}
		
		//modify the file name to get rid of the extension and use this string as the entry name
		String shortName = name;
		if (name.endsWith(".mol2")){
			if (structure.getUrlString() != null){
				shortName = structure.getUrlString();
			}
			else{
				shortName = name.substring(0, name.length()-5);
			}
		}	

		if(shortName.indexOf(File.separator) > -1){
			int index = shortName.lastIndexOf(File.separator);
			String temp = shortName.substring(index+1);
			shortName = temp;
		}

		entry.setStructure(structure);
		entry.setName(shortName);
		
		return entry;
	
	
	}
	

	
	/**
	 * Loads the Entry from a URL object.
	 */
	public Entry load(URL url, int number){
	
		StructureEntry entry = new StructureEntry();
		
		String name = url.toString();
		
		Structure structure = loadStructure( url );

		if ( structure == null ){
			return null;
		}
		
		entry.setSourceUrl(url);
		
		//modify the file name to get rid of the extension and use this string as the entry name
		String shortName = name;
		if (name.endsWith(".mol2")){
			shortName = name.substring(0, name.length()-4);
		}

		if(shortName.indexOf(File.separator) > -1){
			int index = shortName.lastIndexOf(File.separator);
			String temp = shortName.substring(index+1);
			shortName = temp;
		}

		entry.setStructure(structure);
		entry.setName(shortName);
		
		return entry;
	}

	
	/**
	 * Returns true if the loader is capable of loading the structure,
	 * or false otherwise. This enables higher-level code to be able
	 * to build a context sensative menu of only the loaders that can
	 * load a given structure name.
	 */
	public boolean canLoad( String name )
	{
		// System.err.println( "PdbStructureLoader.canLoad(String)" );
		if ( ! name.endsWith( ".mol2" ) )
			return false;

		return true;
	}

	//
	// FileStructureLoader interface methods
	//
	
	/**
	 * Returns a reference to a Structure read from the given File object.
	 */
	public Structure loadStructure( File file )
	{
		// System.err.println( "PdbStructureLoader.load(File)" );
		
//		System.out.println("loadStructure with id = " + identificator);

		if ( ! canLoad( file ) ) return null;
		if ( ! file.exists() ) return null;
		if ( ! file.canRead() ) return null;

		expectedInputBytes = file.length( );
		if ( expectedInputBytes <= 0 ) return null;

		try
		{
			FileInputStream fileInputStream = new FileInputStream( file );
			if ( fileInputStream == null ) return null;
			
			//filename was being reduced to just the name itself here and in SequenceViewerPanel
//			urlString = file.toURL().toExternalForm();
			
			String tempUrlString = null;
			tempUrlString = file.toURL().toExternalForm();
			
			//process the name to get rid of the full path and keep
			//only the file name without the extension for internal reference purposes
			int index = tempUrlString.lastIndexOf("/") + 1;
			String tUrlString = tempUrlString.substring(index);
		
			index = tUrlString.lastIndexOf(".");
			urlString = tUrlString.substring(0,index);
			
			
			BufferedInputStream bufferedInputStream = new BufferedInputStream( fileInputStream );

			return loadStructure( bufferedInputStream );
		}
		catch ( FileNotFoundException e )
		{
			return null;
		}
		catch ( MalformedURLException e )
		{
			return null;
		}
		catch ( SecurityException e )
		{
			return null;
		}
	}


	/**
	 * Returns true if the Entry can be read from the given File object.
	 */
	public boolean canLoad( File file )
	{
		if ( ! file.exists() ) return false;

		try
		{
			// String path = file.getCanonicalPath( );
			String path = file.toURL().toExternalForm();
			return canLoad( path );
		}
		catch( IOException e )
		{
			return false;
		}
	}


	/**
	 * Returns a reference to a Structure read from the given URL object.
	 */
	public Structure loadStructure( URL url )
	{
		// System.err.println( "PdbStructureLoader.load(URL)" );

		if ( ! canLoad( url ) ) return null;

		try
		{
			URLConnection urlConnection = url.openConnection( );
			expectedInputBytes = urlConnection.getContentLength( );
			if ( expectedInputBytes <= 0 ) return null;
			InputStream inputStream = urlConnection.getInputStream( );
			
			if ( inputStream != null ) urlString = url.toExternalForm( );
			BufferedInputStream bufferedInputStream = new BufferedInputStream( inputStream );
			return loadStructure( bufferedInputStream );
		}
		catch( IOException e )
		{
			return null;
		}
	}


	/**
	 * Returns true if the Entry can be read from the given URL object.
	 */
	public boolean canLoad( URL url )
	{
		// System.err.println( "PdbStructureLoader.canLoad(URL)" );
		return canLoad( url.toExternalForm( ) );
	}

	//
	// PdbStructureLoader methods
	//

	/**
	 * Returns a reference to a Structure read from the given (uncompressed)
	 * InputStream.
	 */
	public Structure loadStructure( BufferedInputStream bufferedInputStream )
	{
		// System.err.println( "PdbStructureLoader.load(BufferedInputStream)" );
		if ( bufferedInputStream == null ) return null;

		passComponents = new Hashtable( );
		long expectedBytes = expectedInputBytes;
		// System.err.println( "PdbStructureLoader.load: expectedBytes = " + expectedBytes );

		byte buf[] = new byte[ 2048 ]; // raw read data buffer
		byte line[] = new byte[ 200 ]; // accumulated line buffer

		int expectedReads = (int) ((float) expectedBytes / (float) buf.length);
		// System.err.println( "PdbStructureLoader.load: expectedReads = " + expectedReads );
		int onePercent = (int) (((float) expectedBytes / (float) buf.length) / 100.f);
		if ( onePercent <= 0 ) onePercent = 1;
		// System.err.println( "PdbStructureLoader.load: onePercent = " + onePercent );
		float percentDone = 0.0f;

		Status.progress( percentDone, "Chargement " + urlString );
		
		
		int linePos = 0;
		int bytesRead = 0;
		int lastRead = 0;
		int lines = 0;
		int readCount = 0;
		
		boolean terminate = false;

		boolean atomFlag = false;
		boolean bondFlag = false;
		boolean moleculeFlag = false;
		boolean substructureFlag = false;
		
		boolean substructureRead = false;//once this is true and an @ is found, terminate reading
		
		int tempAtomNum = 0;//used to count atom offsets of residue records
		int tempResNum = 0;
		String tempChainId = null;
		String tempRes = null;
		
		int nameLine = 0;
		String name = null;
		
		int lineNumber = -1;//once it hits the first line, it becomes 0
		
		//variables to handle different formats of bond records
		boolean first = true;
		int startAtom1 = -1;
		int spanAtom1 = -1;
		int startAtom2 = -1;
		int spanAtom2 = -1;
		int startOrder = -1;
		int spanOrder = -1;

		try
		{
			while ( bufferedInputStream.available() > 0 )
			{
				
				if (terminate){
					break;
				}
				
				// Re-fill the raw data buffer.
				try
				{
					lastRead = bufferedInputStream.read( buf, 0, buf.length );
				}
				catch( IOException e )
				{
					System.err.println( "Mol2StructureLoader.load: " + e );
					return null;
				}
				bytesRead += lastRead;

				readCount++;
				if ( readCount % onePercent == 0 )
				{
					percentDone = (float) bytesRead / (float) expectedBytes;
					Status.progress( percentDone, "Chargement " + urlString );
				}
				
				// Process the buffer
				for ( int bufPos=0; bufPos<lastRead; bufPos++ )
				{
					// Copy a byte from the input buffer into the line buffer
					line[linePos] = buf[bufPos];
					if ( buf[bufPos] == '\n' )
					{
						// line[0-linePos] now has a complete line.
						lines++;
						linePos = 0;
						lineNumber++;
						
						
						int atomCount = 0;
						int bondCount = 0;
						
						if (moleculeFlag){
							if (lineNumber == nameLine){
								String temp = (new String(line, 0, 50));
								if (temp.indexOf(".") == -1){
									int n = temp.indexOf("\n");
									if (n < 0){
										name = temp.trim();
									}
									else{
										name = temp.substring(0, n).trim();
									}
								}
								else{
									//it's just a file name
									nameLine += 5;
								}
							}
						}
						
						//check the flags in the line
						if (
							(line[0] == '@') &&
							(line[9] == 'M') &&
							(line[10] == 'O') &&
							(line[11] == 'L') &&
							(line[12] == 'E') &&
							(line[13] == 'C') &&
							(line[14] == 'U') &&
							(line[15] == 'L') &&
							(line[16] == 'E')
						){
							//the molecule flag has started
							//get the atom and bond numbers (later)
//							byte[] line_temp = (byte[])data.get(i);
//							name = new String(line_temp);
//							System.out.println("name = " + name);
							moleculeFlag = true;
							nameLine = lineNumber + 1;
							continue;
						}
							
						if (
							(line[0] == '@') &&
							(line[9] == 'A') &&
							(line[10] == 'T') &&
							(line[11] == 'O') &&
							(line[12] == 'M')
						){
							//the atom flag has started
							atomFlag = true;
							moleculeFlag = false;
							continue;
						}
						
						if (
								(line[0] == '@') &&
								(line[9] == 'B') &&
								(line[10] == 'O') &&
								(line[11] == 'N') &&
								(line[12] == 'D')
						){
							//the molecule flag has started
							bondFlag = true;
							atomFlag = false;
							continue;
						}
						
						
						if (
								(line[0] == '@') &&
								(line[9] == 'S') &&
								(line[10] == 'U') &&
								(line[11] == 'B') &&
								(line[12] == 'S') &&
								(line[13] == 'T') &&
								(line[14] == 'R') &&
								(line[15] == 'U') &&
								(line[16] == 'C')
						){
							//the molecule flag has started
							bondFlag = false;
							substructureFlag = true;
							continue;
						}
						
						if (line[0] == '@' && substructureRead){
							terminate = true;
							break;
						}
						
						if (atomFlag){
							//read the current line and create an atom record
//							System.out.println("Atom found");
							Atom atom = new Atom( );
							try{
	
								atom.number =  Integer.parseInt( (new String( line, 0, 7 )).trim() );
	
//								System.out.print(atom.number + "\n");
								String block = (new String( line, 47, 6 )).trim();
								if (block.equals("Du")){
									atom.element = 0;//mark it as a lone pair element
								}
								else{
									int dot = block.indexOf(".");
									if (dot == -1){
										atom.setElement(block);
									}
									else{
										atom.setElement(block.substring(0, dot));
										if (!PeriodicTable.isElement(atom.element)){
											atom.element = 6;//default to C
										}
									}
								}
								
	//							System.out.print(atom.element + "\t");
								
								atom.name = (new String( line, 8, 4 )).trim();
	//							System.out.print(atom.name + "\t");
								
								try{
									String temp = (new String( line, 59, 8 )).trim();
									temp.replace('.', '_');
									if (temp != null && temp.length() >= name.length() && temp.startsWith(name)){
										atom.compound = name;
									}
									else{
										//it's probably a residue name
										try{
											atom.compound = temp.substring(0,3);
										}
										catch (Exception e){
											atom.compound = name;
										}
									}
								}
								catch (Exception exx){
//									exx.printStackTrace();
									atom.compound = "XXX";
								}
//								System.out.print(atom.compound + "\t");
								
								atom.chain_id = "_";
								
								atom.residue_id = Integer.parseInt((new String( line, 53, 5 )).trim());
//								System.out.print(atom.residue_id + "\t");
								
								atom.partialCharge = Float.parseFloat((new String( line, 69, 7 )).trim());
								
								atom.coordinate = new double[3];
								atom.coordinate[0] =
									Double.parseDouble( (new String( line, 16, 10 )).trim() );
	//							System.out.print(atom.coordinate[0] + "\t");
								atom.coordinate[1] =
									Double.parseDouble( (new String( line, 26, 10 )).trim() );
	//							System.out.print(atom.coordinate[1] + "\t");
								atom.coordinate[2] =
									Double.parseDouble( (new String( line, 36, 10 )).trim() );
	//							System.out.print(atom.coordinate[2] + "\t\t");
	
//								System.out.print(atom.partialCharge + "\n");
								
//								atom.bfactor = 0.0f;
//								atom.occupancy = 1.0f;
								
								//no exceptions encountered: add the new atom 
								Vector records = (Vector) passComponents.get(
										StructureComponentRegistry.TYPE_ATOM );
									if ( records == null )
									{
										records = new Vector( );
										passComponents.put(
											StructureComponentRegistry.TYPE_ATOM, records );
									}
								records.add( atom );
//								System.out.println("added " + atom.name + " to " + records.size());

								
							}
							catch (Exception ex){
								//try reading by simple tokenizing of the string without regard for the 
								//offsets
								try{
									String content = new String(line);
									if (content.indexOf("\n") > -1){
										content = content.substring(0, content.indexOf("\n")).trim();
									}
									StringTokenizer tok = new StringTokenizer(content, " ", false);
									atom.number = Integer.parseInt(tok.nextToken());
									atom.name = tok.nextToken();
									
									atom.coordinate = new double[3];
									atom.coordinate[0] = Double.parseDouble(tok.nextToken() );
									atom.coordinate[1] = Double.parseDouble(tok.nextToken());
									atom.coordinate[2] = Double.parseDouble(tok.nextToken());
									
									String block = tok.nextToken();
									int dot = block.indexOf(".");
									if (dot == -1){
										atom.setElement(block);
									}
									else{
										atom.setElement(block.substring(0, dot));
										if (!PeriodicTable.isElement(atom.element)){
											atom.element = 6;//default to C
										}
									}
									
									atom.residue_id = Integer.parseInt(tok.nextToken());
									
									
									try{
										String temp = tok.nextToken();
										temp.replace('.', '_');
										if (temp != null && temp.length() >= name.length() && temp.startsWith(name)){
											atom.compound = name;
										}
										else{
											//it's probably a residue name
											try{
												atom.compound = temp.substring(0,3);
											}
											catch (Exception e){
												atom.compound = name;
											}
										}
									}
									catch (Exception exx){
//										exx.printStackTrace();
										atom.compound = "XXX";
									}
									
									atom.partialCharge = Float.parseFloat(tok.nextToken());
									
									atom.chain_id = "_";


//									System.out.println("atom  =  " + atom.name + ", charge = " + atom.partialCharge);
									
									//no exceptions encountered: add the new atom 
									Vector records = (Vector) passComponents.get(
											StructureComponentRegistry.TYPE_ATOM );
										if ( records == null )
										{
											records = new Vector( );
											passComponents.put(
												StructureComponentRegistry.TYPE_ATOM, records );
										}
									records.add( atom );
//									System.out.println("added " + atom + " to " + records.size());
								}
								catch (Exception eee){
									continue;
								}
							}
							
							
						}
						
						if (bondFlag){
							
							if (first){
								
								//check for the format of the bond records
								String block = new String(line, 0, 30);
								int endIndex = block.indexOf("\n");
								String sample = block;
								if (endIndex > -1){
									sample = block.substring(0, block.indexOf("\n"));
								}
								
								char[] set = sample.toCharArray();
								for (int i = 0; i < set.length; i++){
									if (i == 0) continue;
									if (set[i] == ' ' && set[i-1] != ' '){
										if (startAtom1 == -1){
											startAtom1 = i;
										}
										else if (startAtom2 == -1){
											startAtom2 = i;
											spanAtom1 = i - startAtom1;
										}
										else if (startOrder == -1){
											startOrder = i+1;
											spanAtom2 = startOrder - startAtom2;
										}
									}
								}
								spanOrder = 2;
								
								first = false;
								
							}
							
							int n1 = (Integer.parseInt(new String(line, startAtom1, spanAtom1).trim())) - 1;
							int n2 = (Integer.parseInt(new String(line, startAtom2, spanAtom2).trim())) - 1;
							
							String order = new String(line, startOrder, spanOrder).trim();
		//					System.out.println("order = #" + order + "#");
							
//							System.out.println("n1 = " + n1 + ", n2 = " + n2);
							
							//get the right atoms from the records
							Atom a1 = (Atom)((Vector) passComponents.get(
									StructureComponentRegistry.TYPE_ATOM )).get(n1);
							Atom a2 = (Atom)((Vector) passComponents.get(
									StructureComponentRegistry.TYPE_ATOM )).get(n2);
							
							Bond bond = new Bond(a1, a2);
							if (order.startsWith("1")){
								bond.setOrderOnly(1.0f);//the default is already SP3 and atoms with another value should not be touched
							}
							else if (order.startsWith("2")){
								bond.setOrderOnly(2.0f);
								a1.hybridization = Atom.SP2;
								a2.hybridization = Atom.SP2;
							}
							else if (order.startsWith("3")){
								bond.setOrderOnly(3.0f);
								a1.hybridization = Atom.SP;
								a2.hybridization = Atom.SP;
							}
							else if (order.startsWith("ar")){
								bond.setOrderOnly(1.5f);
								
								a1.hybridization = Atom.SP2;
								a2.hybridization = Atom.SP2;
							}
							
							Vector records = (Vector) passComponents.get(
									StructureComponentRegistry.TYPE_BOND );
								if ( records == null )
								{
									records = new Vector( );
									passComponents.put(
										StructureComponentRegistry.TYPE_BOND, records );
								}
							records.add( bond );
							
						}
						
						if (substructureFlag){
							if (line[0] == '@'){
								//record the last residue record from the previous run
								Vector atoms = (Vector) passComponents.get(
										StructureComponentRegistry.TYPE_ATOM );
								for (int k = tempAtomNum-1; k < atoms.size(); k++){
									try{
										Atom a = (Atom)atoms.get(k);
										a.chain_id = tempChainId;
										a.residue_id = tempResNum;
										a.compound = tempRes;
//										System.out.println(a.number + ": " + a.name + ", "+ a.compound + "," + a.residue_id + "," + a.chain_id);
									}
									catch (Exception e){
//										System.out.println("exception in mol2 loader = " + e);
										e.printStackTrace();
									}
								}
								terminate = true;
								break;
							}
							
							//read the records from substructure section
/*							try{
								int num = Integer.parseInt(new String(line, 0, 7).trim());
								String temp = new String(line, 22, 6);
								if (temp.indexOf("\n") > -1){
									temp = temp.substring(0, temp.indexOf("\n"));
								}
//								System.out.println("temp = #" + temp + "#");
								int anum = Integer.parseInt(temp.trim());
								String chain = new String( line, 50, 4).trim();
								String res = new String( line, 59, 4).trim();
								
								//this is the first record
								if (tempRes == null){
									tempRes = res;
									tempChainId = chain;
									tempAtomNum = anum;
									tempResNum = num;
									continue;
								}
								Vector atoms = (Vector) passComponents.get(
										StructureComponentRegistry.TYPE_ATOM );
								for (int k = tempAtomNum-1; k < anum-1; k++){
									try{
										Atom a = (Atom)atoms.get(k);
										a.chain_id = tempChainId;
										a.residue_id = tempResNum;
										a.compound = tempRes;
//										System.out.println(a.number + ": " + a.name + ", "+ a.compound + "," + a.residue_id + "," + a.chain_id);
									}
									catch (Exception e){
										e.printStackTrace();
										System.err.println("e = " + e);
									}
								}
								
								tempRes = res;
								tempChainId = chain;
								tempAtomNum = anum;
								tempResNum = num;

								substructureRead = true;
							}
							catch (Exception ex){
//								ex.printStackTrace();
*/								
								//try a different way of parsing the line
								try{
									int num = -1;
									int anum = -1;
									String res = null;
									String tag = null;
									String chain = null;
							
									String templine = new String(line);
									String ttt = templine.substring(0, templine.indexOf("\n"));
									StringTokenizer stk = new StringTokenizer(ttt, " ", false);
									String[] data = new String[7];
									int i = 0;
									while (stk.hasMoreTokens()){
										data[i++] = stk.nextToken();
										if (i == 6) break;
									}
									
									
									num = Integer.parseInt(data[0]);
									String residue = data[1];
									anum = Integer.parseInt(data[2]);
									tag = data[3];
									if (data[4] != null && data[4].length() > 0){
										String t = data[4];//sometimes, this is residue number
									}
									if (data[5] != null && data[5].length() > 0) chain = data[5];
									if (data[6] != null && data[6].length() > 0) res = data[6];
									
									if (res == null){
										//check residue
										res = residue.substring(0, 3);
									}
									
									if (chain == null){
										chain = "_";
									}
									
									//this is the first record
									if (tempRes == null){
										tempRes = res;
										tempChainId = chain;
										tempAtomNum = anum;
										tempResNum = num;
										continue;
									}

									Vector atoms = (Vector) passComponents.get(
											StructureComponentRegistry.TYPE_ATOM );
									for (int k = tempAtomNum-1; k < anum-1; k++){
										try{
											Atom a = (Atom)atoms.get(k);
											a.chain_id = tempChainId;
											a.residue_id = tempResNum;
											a.compound = tempRes;
//											System.out.println(a.number + ": " + a.name + ", "+ a.compound + "," + a.residue_id + "," + a.chain_id);
										}
										catch (Exception e){
											e.printStackTrace();
//											System.err.println("e = " + e);
										}
									}
									
									tempRes = res;
									tempChainId = chain;
									tempAtomNum = anum;
									tempResNum = num;

									substructureRead = true;

								}
								catch (Exception e){
									e.printStackTrace();
								}

							}
							
							
//						}
						
						
						
					}
					else
					{
						// Continue building the line.
						linePos++;
						if ( linePos >= line.length )
						{
//							System.err.println( "line buffer overflow!" );
							return null;
						}
					}
				}
			}
			
			
			
		}
		catch( Exception e )
		{
			e.printStackTrace();
//			System.err.println( "Mol2StructureLoader.load: " + e );
			Status.progress( 1.0f, null );
			return null;
		}
			
		
		if (name != null) urlString = name;
				
//		System.out.println("Urlstring = " + urlString);
				
		//
		// Create the Structure object
		//
		Structure structure = new Structure()
		{
			// A hashtable of vectors where
			// each hash KEY is the StructureComponent type String.
			// each hash VALUE is a Vector of StructureComponent objects.
			protected Hashtable structureComponents = null;

			// To free up the global state for another load call.
			private String localUrlString;

			// public Structure()  Anonymous inner class constructor.
			{
				structureComponents = passComponents;
				passComponents = null;

				localUrlString = urlString;
				urlString = null;
				
			}

			public String getUrlString( ){
				return localUrlString;
			}

			
			public int getStructureComponentCount( String scType ){
				Vector records = (Vector) structureComponents.get( scType );
				if ( records == null )
					return 0;
				else
					return records.size( );
			}

			private void getStructureComponentByIndex( int index,
				StructureComponent structureComponent )
				throws IndexOutOfBoundsException, IllegalArgumentException
			{
				String type = structureComponent.getStructureComponentType();
				if ( type == null )
					throw new IllegalArgumentException( "type inconnu" + type );

				Vector records = (Vector) structureComponents.get( type );
				if ( records == null )
					throw new IllegalArgumentException( "pas d\'enregistrement du type " + type );

				StructureComponent cachedComponent =
					(StructureComponent) records.elementAt( index );

				cachedComponent.structure = this;
				structureComponent.structure = this;
				structureComponent.copy( cachedComponent );
            }

			public StructureComponent getStructureComponentByIndex( String type,
				int index )
				throws IndexOutOfBoundsException, IllegalArgumentException
			{
				Vector records = (Vector) structureComponents.get( type );
				if ( records == null )
					throw new IllegalArgumentException( "pas d\'enregistrement du type " + type );

				StructureComponent structureComponent =
					(StructureComponent) records.elementAt( index );
				structureComponent.structure = this;

				return structureComponent;
            }
		};

		// Progress is done.
		Status.progress( 1.0f, null );
		
		return structure;
	}
	
	public int getDataType(){
		return DATA_TYPE;
	}
		
}

